module Higgs
  use Integrate
  use QCD
  use PDF
  use Evolution
  use HiggsVar
  use HiggsEW
  implicit none

  real(8), private :: gam0, psi0, psi1, psi2, psi3

  real(8), private :: H(0:2), U(0:2), S(0:2,0:4), HUS(0:4)

contains

  subroutine Update_mus(rs)
    implicit none
    real(8), intent(in) :: rs
    real(8) :: mus1, mus2
    real(8) :: a1, b1, c1, a2, b2, c2
    a1 = 0.999d0
    b1 = 0.0721d0
    c1 = 32.47d0
    a2 = 1.098d0
    b2 = 0.344d0
    c2 = 7.184d0
    mus1 = mH * (1d0-tau) / (a1 + b1*tau**0.25d0)**c1
    mus2 = mH * (1d0-tau) / (a2 + b2*tau**0.25d0)**c2
    mu_s = (1d0-rs/2d0) * mus1 + (rs/2d0) * mus2
  end subroutine Update_mus

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This should be called in the beginning and whenever mu's or mH is changed.
  subroutine InitHiggs(threshold)
    implicit none
    logical, intent(in) :: threshold
    complex(8) :: Ax

    tau = (mH/Ecms)**2
    mb = Quark_Mass_MSbar(mH, mbmb, mbmb)

    if (muf0 > 0d0) then
       mu_f = muf0
    else
       mu_f = -muf0*mH
    end if

    if (threshold) then
       if (mut0 > 0d0) then
          mu_t = mut0
       else
          mu_t = -mut0*mt
       end if
       if (muh0 > 0d0) then
          mu_h = muh0
       else
          mu_h = -muh0*mH
       end if
       if (mus0 > 0d0) then
          mu_s = mus0
       else
          call Update_mus(-mus0)
       end if
    else
       mu_t = mu_f
       mu_h = mu_f
       mu_s = mu_f
    end if

    Lt = 2d0*log(mt/mu_t)
    Ls = 2d0*log(mH/mu_s)
    alphas_mut = Get_Alphas(mu_t)
    alphas_muh = Get_Alphas(mu_h)
    alphas_imuh = Alphas_Lambda_Complex(dcmplx(-mu_h**2,-EPS))
    alphas_mus = Get_Alphas(mu_s)
    alphas_muf = Get_Alphas(mu_f)
    call Update_eta(gammaCuspA)

    Ax = A(4*mt**2/mH**2) + A(4*mb**2/mH**2)
    delta_EW = delta_EW_Spline()
    prefactor = hbar_c2 * (Ax * conjg(Ax)) * GFew / (288d0 * sqrt2 * pi) * tau
    prefactor2 = hbar_c2 * (Ax * conjg(Ax)) * GFew / (288d0 * sqrt2) * pi
  end subroutine InitHiggs

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine Higgs_Resummed_Integrand(ndim, y, ncomp, r)
    implicit none
    integer, intent(in) :: ndim, ncomp
    real(8), intent(in) :: y(ndim)
    real(8), intent(out) :: r(ncomp)
    real(8) :: z, x
    real(8) :: pdf(-6:6), fxx, fz, ft(0:2)
    real(8) :: d_eta(0:4)
    integer :: i

    x = tau + (1d0-tau)*y(1)
    z = tau/x + (1d0-tau/x)*y(2)

    ! fxx = f(x)/x
    call evolvePDF(x, mu_f, pdf)
    fxx = pdf(0)/x**2

    ! fz = f(tau/x/z)
    call evolvePDF(tau/x/z, mu_f, pdf)
    fz = pdf(0)*z*x/tau

    ! ft(0) = f(tau/x), ft(1) = f'(tau/x), ft(2) = f''(tau/x)
    call fg_gen(tau/x, sub_order, ft)

    call Higgs_subtraction(z, fz, tau/x, ft, d_eta)

    r = 0d0
    do i = 0, 4
       r = r + HUS(i)*d_eta(i)
    end do

    r = r * prefactor * (1d0-tau) * (1d0-tau/x) * fxx
  end subroutine Higgs_Resummed_Integrand

  function Higgs_Resummed_CS(threshold, iPi) result (cs)
    implicit none
    logical, intent(in) :: threshold, iPi
    real(8) :: cs, r(1)

    call InitHiggs(threshold)

    call Higgs_HUS(HUS, iPi)

    call Cuba(Cuba_CUHRE, 2, 1, Higgs_Resummed_Integrand, r, iERR)
    cs = r(1)
  end function Higgs_Resummed_CS

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine Higgs_Leading_Integrand(ndim, y, ncomp, r)
    implicit none
    integer, intent(in) :: ndim, ncomp
    real(8), intent(in) :: y(ndim)
    real(8), intent(out) :: r(ncomp)
    real(8) :: z, x
    real(8) :: l1z, plus, delta
    real(8) :: pdf(-6:6), fxx, fzz, ft

    x = tau + (1d0-tau)*y(1)
    z = y(2)

    ! fxx = f(x)/x
    call evolvePDF(x, mu_f, pdf)
    fxx = pdf(0)/x**2

    ! fzz = f(tau/x/z)/z
    if (z > tau/x) then
       call evolvePDF(tau/x/z, mu_f, pdf)
       fzz = pdf(0)*x/tau
    else
       fzz = 0d0
    end if

    ! ft = f(tau/x)
    call evolvePDF(tau/x, mu_f, pdf)
    ft = pdf(0)*x/tau

    l1z = log(mH**2*(1d0-z)**2/mu_f**2/z)
    delta = 1d0
    plus = 0d0
    if (higgs_order > 0) then
       delta = delta + (alphas_muf/pi) * 25.23921d0
       plus = plus + (alphas_muf/pi) * (6d0*l1z)
    end if
    if (higgs_order > 1) then
       delta = delta + (alphas_muf/pi)**2 * ( &
            -4.302024d0*2d0*log(mH/mu_f) - 5.708333d0*2d0*log(mt/mu_f) + 338.5086d0 )
       plus = plus + (alphas_muf/pi)**2 * (9d0*l1z**3 - 5.75d0*l1z**2 + &
            (-15.85537d0 - 9d0*4d0*log(mH/mu_f)**2) * l1z + 222.9056d0)
    end if
    r = plus * fxx * (fzz - ft) / (1d0-z) + delta * fxx * ft
    r = r * prefactor * (1d0-tau) * alphas_muf**2
  end subroutine Higgs_Leading_Integrand

  function Higgs_Leading_CS() result (cs)
    implicit none
    real(8) :: cs, r(1)

    call InitHiggs(.false.)

    call Cuba(Cuba_CUHRE, 2, 1, Higgs_Leading_Integrand, r, iERR)
    cs = r(1)
  end function Higgs_Leading_CS
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine Update_eta(gcusp)
    implicit none
    real(8), intent(in) :: gcusp(0:3)
    complex(8) :: aGsf(0:2)

!!!!! We need to decide to use Res_a or Res_aG here !!!!!
    call Res_a(dcmplx(alphas_mus), dcmplx(alphas_muf), gcusp, beta, 2, aGsf)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    eta = dble(2d0*(aGsf(0) + aGsf(1) + aGsf(2)))

    gam0 = exp(-2d0*gammaE*eta)/dgammf(1d0+2d0*eta)
    psi0 = dpsipg(1d0+2d0*eta,0)
    psi1 = dpsipg(1d0+2d0*eta,1)
    psi2 = dpsipg(1d0+2d0*eta,2)
    psi3 = dpsipg(1d0+2d0*eta,3)
  end subroutine Update_eta

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Implement the star-subtraction

  subroutine Higgs_Subtraction(z, fz, t, ft, d_eta)
    implicit none
    real(8), intent(in) :: z, fz, t, ft(0:2)
    real(8), intent(out) :: d_eta(0:4)
    real(8) :: feta, l1zg, l1zg0, l1zg1, l1zg2, l1zg3, lz, sub(3,0:4)
    real(8) :: l1taug, l1taug0, l1taug1, l1taug2, l1taug3, add(3,0:4)

    feta = z**(-1d0-eta) * (1d0-z)**(-1d0+2d0*eta) * gam0 * fz
    sub(1,0) = (1d0-z)**(-1d0+2d0*eta) * gam0 * ft(0) * (2d0*eta)
    add(1,0) = (1d0-t)**(-1d0+2d0*eta) * gam0 * ft(0)
    d_eta(0) = 2d0*eta*feta - sub(1,0) + add(1,0)

    if (higgs_order > 0) then
       lz = log(z)

       l1zg = log(1d0-z) - gammaE
       l1zg0 = l1zg - psi0
       l1zg1 = l1zg0**2 - psi1
       sub(1,1) = (1d0-z)**(-1d0+2d0*eta) * gam0 * ft(0) &
            * (4d0*l1zg0*eta + 2d0)
       sub(1,2) = (1d0-z)**(-1d0+2d0*eta) * gam0 * ft(0) &
            * 8d0*(eta*l1zg1 + l1zg0)

       l1taug = log(1d0-t) - gammaE
       l1taug0 = l1taug - psi0
       l1taug1 = l1taug0**2 - psi1
       add(1,1) = add(1,0) * (2d0*l1taug0)
       add(1,2) = add(1,0) * (4d0*l1taug1)

       d_eta(1) = feta * (4d0*l1zg0*eta - 2d0*lz*eta + 2d0) &
            - sub(1,1) + add(1,1)
       d_eta(2) = feta * &
            ( 2d0*eta*(4d0*l1zg1+lz**2) + 8d0*(1d0-eta*lz)*l1zg0 - 4d0*lz ) &
            - sub(1,2) + add(1,2)
    end if

    if (higgs_order > 1) then
       l1zg2 = -2d0*l1zg0**3 + 3d0*l1zg0*l1zg1 - psi2
       l1zg3 = 6d0*l1zg0**4 - 12d0*l1zg0**2*l1zg1 + 4d0*l1zg0*l1zg2 + 3d0*l1zg1**2 - psi3
       sub(1,3) = (1d0-z)**(-1d0+2d0*eta) * gam0 * ft(0) &
            * 8*(3*l1zg1 + 2*eta*l1zg2)
       sub(1,4) = (1d0-z)**(-1d0+2d0*eta) * gam0 * ft(0) &
            * 32*(2*l1zg2 + eta*l1zg3)

       l1taug2 = -2d0*l1taug0**3 + 3d0*l1taug0*l1taug1 - psi2
       l1taug3 = 6d0*l1taug0**4 - 12d0*l1taug0**2*l1taug1 + 4d0*l1taug0*l1taug2 + 3d0*l1taug1**2 - psi3
       add(1,3) = add(1,0) * (8d0*l1taug2)
       add(1,4) = add(1,0) * (16d0*l1taug3)

       d_eta(3) = feta * &
            ( 8*(3*l1zg1 + 2*eta*l1zg2) - 24*(l1zg0 + eta*l1zg1)*lz + &
            6*(1 + 2*eta*l1zg0)*lz**2 - 2*eta*lz**3 ) &
            - sub(1,3) + add(1,3)
       d_eta(4) = feta * &
            ( 32*eta*l1zg3 - 96*l1zg1*lz + 48*(l1zg0 + eta*l1zg1)*lz**2 - &
            8*(1 + 2*eta*l1zg0)*lz**3 + 2*eta*lz**4 - 64*l1zg2*(-1 + eta*lz) ) &
            - sub(1,4) + add(1,4)
    end if

    if (sub_order > 1) then
       sub(2,0) = (1d0-z)**(2d0*eta) * gam0 * 2d0*eta*((1d0+eta)*ft(0) + t*ft(1))
       add(2,0) = (1d0-t)**(2d0*eta) * gam0 * (2d0*eta) / (1d0+2d0*eta) * &
            ( (1d0+eta)*ft(0) + t*ft(1) )
       d_eta(0) = d_eta(0) - sub(2,0) + add(2,0)

       if (higgs_order > 0) then
          sub(2,1) = (1d0-z)**(2d0*eta) * gam0 * 2d0 * &
               ( ft(0) * (2d0*eta*(l1zg0*(1d0+eta)+1d0) + 1d0) &
               + ft(1) * t * (l1zg0*2d0*eta + 1d0) )
          sub(2,2) = (1d0-z)**(2d0*eta) * gam0 * 4d0 * &
               ( ft(0) * (l1zg1*2d0*eta*(1d0+eta) + l1zg0*2d0*(1d0+2d0*eta) + 1d0) &
               + ft(1) * 2d0*t * (l1zg1*eta + l1zg0) )

          add(2,1) = (1d0-t)**(2d0*eta) * gam0 / (1d0+2d0*eta)**2 * 2d0 * &
               ( ft(0) * (2d0*eta*(1d0+eta)*(l1taug0*(1d0+2d0*eta)+1d0) + 1d0) &
               + ft(1) * t * (l1taug0*2d0*eta*(1d0+2d0*eta) + 1d0) )
          add(2,2) = (1d0-t)**(2d0*eta) * gam0 / (1d0+2d0*eta)**3 * 4d0 * &
               ( ft(0) * (2d0*(2d0*eta*(eta*(2d0*eta+3d0)+2d0)+1d0)*l1taug0 + &
               2d0*eta*(1d0+eta)*(1d0+2d0*eta)**2*l1taug1 - 1d0) &
               + ft(1) * 2d0*t*(eta*(1d0+2d0*eta)**2*l1taug1 + &
               (1d0+2d0*eta)*l1taug0 - 1d0) )

          d_eta(1) = d_eta(1) - sub(2,1) + add(2,1)
          d_eta(2) = d_eta(2) - sub(2,2) + add(2,2)
       end if

       if (higgs_order > 1) then
          sub(2,3) = (1d0-z)**(2d0*eta) * gam0 * &
               ( 8*((3*(l1zg0 + l1zg1 + 2*eta*l1zg1) + 2*eta*(1 + eta)*l1zg2)*ft(0) + &
               (3*l1zg1 + 2*eta*l1zg2)*t*ft(1)) )
          sub(2,4) = (1d0-z)**(2d0*eta) * gam0 * &
               (  32*((3*l1zg1 + (2 + 4*eta)*l1zg2 + eta*(1 + eta)*l1zg3)*ft(0) + &
               (2*l1zg2 + eta*l1zg3)*t*ft(1)) )

          add(2,3) = (1d0-t)**(2d0*eta) * gam0 / (1d0+2d0*eta)**4 * &
               ( 8*((-3*(1 + 2*eta)*l1taug0 + 3*(1 + l1taug1) + 2*eta*(1 + eta)* &
               (3*(3 + 4*eta*(1 + eta))*l1taug1 + (1 + 2*eta)**3*l1taug2))*ft(0) + &
               (6 - 6*(1 + 2*eta)*l1taug0 + 3*(1 + 2*eta)**2*l1taug1 + 2*eta*(1 + 2*eta)**3*l1taug2)*t*ft(1)) )
          add(2,4) = (1d0-t)**(2d0*eta) * gam0 / (1d0+2d0*eta)**5 * &
               ( 32*((-6 + 6*(1 + 2*eta)*l1taug0 - 3*(1 + 2*eta)**2*l1taug1 + &
               2*(1 + 2*eta)**3*(1 + 2*eta*(1 + eta))*l1taug2 + &
               eta*(1 + eta)*(1 + 2*eta)**4*l1taug3)*ft(0) + (12*(-1 + l1taug0 + 2*eta*l1taug0) - &
               6*(1 + 2*eta)**2*l1taug1 + 2*(1 + 2*eta)**3*l1taug2 + eta*(1 + 2*eta)**4*l1taug3)*t*ft(1)) )

          d_eta(3) = d_eta(3) - sub(2,3) + add(2,3)
          d_eta(4) = d_eta(4) - sub(2,4) + add(2,4)
       end if
    end if

    if (sub_order > 2) then
       sub(3,0) = (1d0-z)**(1d0+2d0*eta) * gam0 * eta * &
            ( (1d0+eta)*(2d0+eta)*ft(0) + 2d0*t*(2d0+eta)*ft(1) + t**2*ft(2) )
       add(3,0) = (1d0-t)**(1d0+2d0*eta) * gam0 * eta / 2d0 / (1d0+eta) * &
            ( (1d0+eta)*(2d0+eta)*ft(0) + 2d0*t*(2d0+eta)*ft(1) + t**2*ft(2) )
       d_eta(0) = d_eta(0) - sub(3,0) + add(3,0)

       if (higgs_order > 0) then
          sub(3,1) = (1d0-z)**(1d0+2d0*eta) * gam0 * &
               ( ft(0) * (eta*(2d0+eta)*(2d0*(1d0+eta)*l1zg0+3d0)+2d0) &
               + 4d0*ft(1) * t * (eta*(2d0+eta)*l1zg0+1d0+eta) &
               + ft(2) * t**2 * (2d0*eta*l1zg0+1d0) )
          sub(3,2) = (1d0-z)**(1d0+2d0*eta) * gam0 * &
               ( 2d0*ft(0) * (eta*(6d0*(2d0+eta)*l1zg0+2d0*(1d0+eta)*(2d0+eta)*l1zg1 + 3d0) &
               + 4d0*l1zg0 + 3d0) &
               + 4d0*ft(1) * t*(4d0*(1d0+eta)*l1zg0 + 2d0*eta*(2d0+eta)*l1zg1 + 1d0) &
               + 4d0*ft(2) * t**2*(eta*l1zg1 + l1zg0) )

          add(3,1) = (1d0-t)**(1d0+2d0*eta) * gam0 / 2d0 / (1d0+eta)**2 * &
               ( ft(0) * 2d0*(1d0+eta)**2*(eta*(2d0+eta)*l1taug0 + eta + 1d0) &
               + ft(1) * 2d0*t*(eta*(2d0+eta)*(2d0*(1d0+eta)*l1taug0+1d0) + 2d0) &
               + ft(2) * t**2*(2d0*eta*(1d0+eta)*l1taug0 + 1d0) )
          add(3,2) = (1d0-t)**(1d0+2d0*eta) * gam0 / (1d0+eta)**3 * &
               ( ft(0) * (1d0+eta)**3*(4d0*(1d0+eta)*l1taug0 + 2d0*eta*(2d0+eta)*l1taug1 + 1d0) &
               + ft(1) * 2d0*t*(2d0*(1d0+eta)*((eta*(2d0+eta)+2d0)*l1taug0 + &
               eta*(1d0+eta)*(2d0+eta)*l1taug1) - 1d0) &
               + ft(2) * t**2*(2d0*(1d0+eta)*(eta*(1d0+eta)*l1taug1 + l1taug0) - 1d0) )

          d_eta(1) = d_eta(1) - sub(3,1) + add(3,1)
          d_eta(2) = d_eta(2) - sub(3,2) + add(3,2)
       end if

       if (higgs_order > 1) then
          sub(3,3) = (1d0-z)**(1d0+2d0*eta) * gam0 * &
               ( 2*((3 + 18*(1 + eta)*l1zg0 + 12*l1zg1 + 2*eta*(2 + eta)*(9*l1zg1 + 2*(1 + eta)*l1zg2))* ft(0) + &
               2*t*( 2*(3*l1zg0 + 6*(1 + eta)*l1zg1 + 2*eta*(2 + eta)*l1zg2)*ft(1) + &
               (3*l1zg1 + 2*eta*l1zg2)*t*ft(2))) )
          sub(3,4) = (1d0-z)**(1d0+2d0*eta) * gam0 * &
               ( 16*(3*l1zg0*ft(0) + 9*(1 + eta)*l1zg1*ft(0) + 4*l1zg2*ft(0) + 12*eta*l1zg2*ft(0) + &
               6*eta**2*l1zg2*ft(0) + 2*eta*l1zg3*ft(0) + 3*eta**2*l1zg3*ft(0) + eta**3*l1zg3*ft(0) + &
               6*l1zg1*t*ft(1) + 8*l1zg2*t*ft(1) + 8*eta*l1zg2*t*ft(1) + 4*eta*l1zg3*t*ft(1) + &
               2*eta**2*l1zg3*t*ft(1) + (2*l1zg2 + eta*l1zg3)*t**2*ft(2)) )

          add(3,3) = (1d0-t)**(1d0+2d0*eta) * gam0 / (1d0+eta)**4 * &
               ( 6*t*ft(1) + 4*eta*(1 + eta)**3*(2 + eta)*l1taug2* (ft(0) + eta*ft(0) + 2*t*ft(1)) + &
               (3 + 4*eta*(1 + eta)**3*l1taug2)*t**2*ft(2) + &
               6*(1 + eta)**2*l1taug1* (2*(1 + eta)**3*ft(0) + 2*(2 + eta*(2 + eta))*t*ft(1) + t**2*ft(2)) + &
               6*(1 + eta)*l1taug0* ((1 + eta)**3*ft(0) - t*(2*ft(1) + t*ft(2))) )
          add(3,4) = (1d0-t)**(1d0+2d0*eta) * gam0 / (1d0+eta)**5 * &
               ( 4*(2*eta*(1 + eta)**5*(2 + eta)*l1taug3*ft(0) + 2*(-3 + 6*(1 + eta)*l1taug0 + &
               2*eta*(1 + eta)**4*(2 + eta)*l1taug3)*t*ft(1) + (-3 + 6*(1 + eta)*l1taug0 + &
               2*eta*(1 + eta)**4*l1taug3)*t**2*ft(2) + &
               4*(1 + eta)**3*l1taug2* (2*(1 + eta)**3*ft(0) + 2*(2 + eta*(2 + eta))*t*ft(1) + t**2*ft(2)) &
               + 6*(1 + eta)**2*l1taug1* ((1 + eta)**3*ft(0) - t*(2*ft(1) + t*ft(2)))) )

          d_eta(3) = d_eta(3) - sub(3,3) + add(3,3)
          d_eta(4) = d_eta(4) - sub(3,4) + add(3,4)
       end if
    end if
  end subroutine Higgs_Subtraction

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! The hard and soft functions, as well as the evolution factor

  subroutine Higgs_C_t(C_t)
    implicit none
    real(8), intent(out) :: C_t(1:2)
    C_t = 0d0
    if (higgs_order > 0) then
       C_t(1) = (11d0 - beta(1)/beta(0)) * alphas_mut/4d0/pi &
            + beta(1)/beta(0) * alphas_muh/4d0/pi
    endif
    if (higgs_order > 1) then
       C_t(2) = ( 154.2778d0 - 19d0 * Lt + nf * (-11.16667d0 - 5.333333d0 * Lt) &
            - 11d0 * beta(1)/beta(0) + beta(1)**2/beta(0)**2 &
            - beta(2)/beta(0) ) * (alphas_mut/4d0/pi)**2 &
            + beta(2)/beta(0) * (alphas_muh/4d0/pi)**2 &
            + (11d0 * beta(1)/beta(0) - beta(1)**2/beta(0)**2) &
            * (alphas_muh/4d0/pi) * (alphas_mut/4d0/pi)
    endif
  end subroutine Higgs_C_t

  subroutine Higgs_C_S(C_S, iPi)
    implicit none
    complex(8), intent(out) :: C_S(1:2)
    logical, intent(in) :: iPi
    complex(8) :: Lh, alphas
    if (iPi) then
       Lh = 2d0 * log(mH/mu_h)
       alphas = alphas_imuh
    else
       Lh = 2d0 * log(mH/mu_h) - (0d0, pi)
       alphas = alphas_muh
    endif
    C_S = 0d0
    if (higgs_order > 0) then
       C_S(1) = (4.934802d0 - 3d0*Lh**2) * alphas/4d0/pi
    endif
    if (higgs_order > 1) then
       C_S(2) =  ( 4.5d0 * Lh**4 + Lh**3 * (11d0 - 0.666667d0 *nf) &
            + Lh**2 * (-52.1956d0 + 3.333333d0 * nf) + Lh * (-103.536d0 + 15.02418d0 * nf) &
            + 289.209d0 - 62.6491d0 * nf ) * (alphas/4d0/pi)**2
    endif
  end subroutine Higgs_C_S

  subroutine Higgs_H(H, iPi)
    implicit none
    real(8), intent(out) :: H(0:2)
    logical, intent(in) :: iPi
    complex(8) :: C_S(1:2)
    real(8) :: C_t(1:2)
    H = 0d0
    call Higgs_C_t(C_t)
    call Higgs_C_S(C_S, iPi)
    H(0) = 1d0
    if (higgs_order > 0) then
       H(1) = 2d0 * (C_t(1) + dble(C_S(1)))
    endif
    if (higgs_order > 1) then
       H(2) = 2d0 * (C_t(2) + dble(C_S(2))) + C_t(1)**2 + abs(C_S(1))**2 &
            + 4d0 * C_t(1) * dble(C_S(1))
    endif
    H = H * alphas_muh**2
  end subroutine Higgs_H

  subroutine Higgs_U(U, iPi)
    implicit none
    real(8), intent(out) :: U(0:2)
    logical, intent(in) :: iPi
    real(8) :: Uexp(0:2), L
    complex(8) :: Shs(0:2), aGhs(0:2), aVhs(0:2), aPsf(0:2)
    complex(8) :: SiPi(0:2), aGiPi(0:2), aViPi(0:2), UiPi(0:2)

    L = log(mH/mu_h)
    call Res_S(dcmplx(alphas_muh), dcmplx(alphas_mus), gammaCuspA, beta, higgs_order, Shs)
    call Res_a(dcmplx(alphas_muh), dcmplx(alphas_mus), gammaCuspA, beta, higgs_order, aGhs)
    call Res_a(dcmplx(alphas_muh), dcmplx(alphas_mus), gamma_VH, beta, higgs_order, aVhs)
    call Res_a(dcmplx(alphas_mus), dcmplx(alphas_muf), gamma_phi, beta, higgs_order, aPsf)
    Uexp = dble(4d0 * Shs - 4d0 * L * aGhs - 2d0 * aVhs + 4d0 * aPsf)

    if (iPi) then
       call Res_S(alphas_imuh, dcmplx(alphas_muh), gammaCuspA, beta, Higgs_order, SiPi)
       call Res_a(alphas_imuh, dcmplx(alphas_muh), gammaCuspA, beta, Higgs_order, aGiPi)
       call Res_a(alphas_imuh, dcmplx(alphas_muh), gamma_V, beta, Higgs_order, aViPi)
       UiPi = 2d0 * SiPi - 2d0 * L * aGiPi - aViPi
       Uexp = Uexp + 2d0 * dble(UiPi)
    end if

    U(0) = exp(Uexp(0))
    U(1) = U(0)*Uexp(1)
    U(2) = U(0)*(Uexp(1)**2 + 2d0*Uexp(2))/2d0
  end subroutine Higgs_U

  subroutine Higgs_S(S)
    implicit none
    real(8), intent(out) :: S(0:2,0:4)

    S = 0d0
    S(0,0) = 1d0
    if (higgs_order > 0) then
       S(1,0) = 6d0*Ls**2 + 9.869604d0
       S(1,1) = 12d0*Ls
       S(1,2) = 6d0
       S(1,:) = S(1,:) * (alphas_mus/4d0/pi)
    endif
    if (higgs_order > 1) then
       S(2,0) = 29.87778d0 + 95.80723d0*Ls + 100.6667d0*Ls**2 - 15.33333d0*Ls**3 + 18d0*Ls**4
       S(2,1) = 95.80723d0 + 201.3333d0*Ls - 46d0*Ls**2 + 72d0*Ls**3
       S(2,2) = 100.6667d0 - 46d0*Ls + 108d0*Ls**2
       S(2,3) = 72d0*Ls - 15.33333d0
       S(2,4) = 18d0
       S(2,:) = S(2,:) * (alphas_mus/4d0/pi)**2
    endif
  end subroutine Higgs_S

  subroutine Higgs_HUS(HUS, iPi)
    implicit none
    real(8), intent(out) :: HUS(0:4)
    logical, intent(in) :: iPi
    call Higgs_H(H, iPi)
    call Higgs_U(U, iPi)
    call Higgs_S(S)
    HUS(:) = H(0)*U(0)*S(0,:)
    if (higgs_order > 0) then
       HUS(:) = HUS(:) + &
            H(1)*U(0)*S(0,:) + H(0)*U(1)*S(0,:) + H(0)*U(0)*S(1,:)
    end if
    if (higgs_order > 1) then
       HUS(:) = HUS(:) + &
            H(2)*U(0)*S(0,:) + H(0)*U(2)*S(0,:) + H(0)*U(0)*S(2,:) + &
            H(1)*U(1)*S(0,:) + H(1)*U(0)*S(1,:) + H(0)*U(1)*S(1,:)
    end if
  end subroutine Higgs_HUS

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

end module Higgs
